const pageResults = require('graph-results-pager');

const { request, gql } = require('graphql-request');

const { graphAPIEndpoints, merchantAddress, TWENTY_FOUR_HOURS } = require('./../constants')
const { timestampToBlock, getAverageBlockTime } = require('./../utils');

const { pairs: exchangePairs } = require('./exchange');
const { priceUSD: innaiPriceUSD } = require('./innai');

module.exports = {
    async info({block = undefined, timestamp = undefined} = {}) {
        block = block ? block : timestamp ? (await timestampToBlock(timestamp)) : undefined;
        block = block ? `block: { number: ${block} }` : "";

        const result = await request(graphAPIEndpoints.mastermerchant,
            gql`{
                    masterMerchant(id: "${merchantAddress}", ${block}) {
                        ${info.properties.toString()}
                    }
                }`
        );

        return info.callback(result.masterMerchant);
    },

    async pool({block = undefined, timestamp = undefined, pool_id = undefined, pool_address = undefined} = {}) {
        if(!pool_id && !pool_address) { throw new Error("genova-data: Pool ID / Address undefined"); }

        block = block ? block : timestamp ? (await timestampToBlock(timestamp)) : undefined;
        block = block ? `block: { number: ${block} }` : "";

        let result;
        if(pool_id) {
            result = await request(graphAPIEndpoints.mastermerchant,
                gql`{
                        pool(id: ${pool_id}, ${block}) {
                            ${pools.properties.toString()}
                        }
                    }`
            );
        }

        else {
            result = await request(graphAPIEndpoints.mastermerchant,
                gql`{
                        pools(first: 1, where: {pair: "${pool_address.toLowerCase()}"}, ${block}) {
                            ${pools.properties.toString()}
                        }
                    }`
            );
        }

        return pools.callback(pool_id ? [result.pool] : result.pools)[0];
    },

    async pools({block = undefined, timestamp = undefined} = {}) {
        return pageResults({
            api: graphAPIEndpoints.mastermerchant,
            query: {
                entity: 'pools',
                selection: {
                    block: block ? { number: block } : timestamp ? { number: await timestampToBlock(timestamp) } : undefined,
                },
                properties: pools.properties
            }
        })
            .then(results => pools.callback(results))
            .catch(err => console.log(err));
    },

    async stakedValue({block = undefined, timestamp = undefined, token_address = undefined} = {}) {
        if(!token_address) { throw new Error("genova-data: Token address undefined"); }

        block = block ? block : timestamp ? (await timestampToBlock(timestamp)) : undefined;
        block = block ? `block: { number: ${block} }` : "";

        const result = await request(graphAPIEndpoints.exchange,
            gql`{
                    liquidityPosition(id: "${token_address.toLowerCase()}-${merchantAddress}", ${block}) {
                        ${stakedValue.properties.toString()}
                    }
                }`
        );

        return stakedValue.callback(result.liquidityPosition);
    },

    async user({block = undefined, timestamp = undefined, user_address = undefined} = {}) {
        if(!user_address) { throw new Error("genova-data: User address undefined"); }

        return pageResults({
            api: graphAPIEndpoints.mastermerchant,
            query: {
                entity: 'users',
                selection: {
                    where: {
                        address: `\\"${user_address.toLowerCase()}\\"`
                    },
                    block: block ? { number: block } : timestamp ? { number: await timestampToBlock(timestamp) } : undefined,
                },
                properties: user.properties
            }
        })
            .then(results => user.callback(results))
            .catch(err => console.log(err));
    },

    async users({block = undefined, timestamp = undefined} = {}) {
        return pageResults({
            api: graphAPIEndpoints.mastermerchant,
            query: {
                entity: 'users',
                selection: {
                    block: block ? { number: block } : timestamp ? { number: await timestampToBlock(timestamp) } : undefined,
                },
                properties: user.properties
            }
        })
            .then(results => user.callback(results))
            .catch(err => console.log(err));
    },

    async apys({block = undefined, timestamp = undefined} = {}) {
        const mastermerchantList = await module.exports.pools({block, timestamp});
        const exchangeList = await exchangePairs({block, timestamp});
        const innaiUSD = await innaiPriceUSD({block, timestamp});

        const totalAllocPoint = mastermerchantList.reduce((a, b) => a + b.allocPoint, 0);

        const averageBlockTime = await getAverageBlockTime({block, timestamp});

        return mastermerchantList.map(mastermerchantPool => {
            const exchangePool = exchangeList.find(e => e.id === mastermerchantPool.pair);
            if(!exchangePool) {
                return {...mastermerchantPool, apy: 0};
            }

            const tvl = mastermerchantPool.innlpBalance * (exchangePool.reserveUSD / exchangePool.totalSupply);
            const innaiPerBlock = (mastermerchantPool.allocPoint / (totalAllocPoint) * 100);
            const apy = innaiUSD * (innaiPerBlock * (60 / averageBlockTime) * 60 * 24 * 365) / tvl * 100;

            return {...mastermerchantPool, apy};
        });
    },

    async apys24h({block = undefined, timestamp = undefined} = {}) {
        let timestampNow = timestamp ? timestamp : block ? await blockToTimestamp(block) : (Math.floor(Date.now() / 1000));
        timestamp24ago = timestampNow - TWENTY_FOUR_HOURS;
        timestamp48ago = timestamp24ago - TWENTY_FOUR_HOURS;

        block = timestamp ? await timestampToBlock(timestamp) : block;
        block24ago = await timestampToBlock(timestamp24ago);
        block48ago = await timestampToBlock(timestamp48ago);

        const results = await module.exports.apys({block: block});
        const results24ago = await module.exports.apys({block: block24ago});

        return apys.callback24h(results, results24ago);
    }
}

const info = {
    properties: [
        'bonusMultiplier',
        'bonusEndBlock',
        'devaddr',
        'migrator',
        'owner',
        'startBlock',
        'innai',
        'innaiPerBlock',
        'totalAllocPoint',
        'poolCount',
        'innlpBalance',
        'innlpAge',
        'innlpAgeRemoved',
        'innlpDeposited',
        'innlpWithdrawn',
        'updatedAt'
    ],

    callback(results) {
        return ({
            bonusMultiplier: Number(results.bonusMultiplier),
            bonusEndBlock: Number(results.bonusEndBlock),
            devaddr: results.devaddr,
            migrator: results.migrator,
            owner: results.owner,
            startBlock: Number(results.startBlock),
            innaiPerBlock: results.innaiPerBlock / 1e18,
            totalAllocPoint: Number(results.totalAllocPoint),
            poolCount: Number(results.poolCount),
            innlpBalance: Number(results.innlpBalance),
            innlpAge: Number(results.innlpAge),
            innlpAgeRemoved: Number(results.innlpAgeRemoved),
            innlpDeposited: Number(results.innlpDeposited),
            innlpWithdrawn: Number(results.innlpWithdrawn),
            updatedAt: Number(results.updatedAt)
        });
    }
};

const pools = {
    properties: [
        'id',
        'pair',
        'allocPoint',
        'lastRewardBlock',
        'accInnovaiPerShare',
        'balance',
        'userCount',
        'innlpBalance',
        'innlpAge',
        'innlpAgeRemoved',
        'innlpDeposited',
        'innlpWithdrawn',
        'timestamp',
        'block',
        'updatedAt',
        'entryUSD',
        'exitUSD',
        'innaiHarvested',
        'innaiHarvestedUSD'
    ],

    callback(results) {
        return results.map(({ id, pair, allocPoint, lastRewardBlock, accInnovaiPerShare, balance, userCount, innlpBalance, innlpAge, innlpAgeRemoved, innlpDeposited, innlpWithdrawn, timestamp, block, updatedAt, entryUSD, exitUSD, innaiHarvested, innaiHarvestedUSD }) => ({
            id: Number(id),
            pair: pair,
            allocPoint: Number(allocPoint),
            lastRewardBlock: Number(lastRewardBlock),
            accInnovaiPerShare: BigInt(accInnovaiPerShare),
            userCount: Number(userCount),
            innlpBalance: Number(innlpBalance),
            innlpAge: Number(innlpAge),
            innlpAgeRemoved: Number(innlpAgeRemoved),
            innlpDeposited: Number(innlpDeposited),
            innlpWithdrawn: Number(innlpWithdrawn),
            addedTs: Number(timestamp),
            addedDate: new Date(timestamp * 1000),
            addedBlock: Number(block),
            lastUpdatedTs: Number(updatedAt),
            lastUpdatedDate: new Date(updatedAt * 1000),
            entryUSD: Number(entryUSD),
            exitUSD: Number(exitUSD),
            innaiHarvested: Number(innaiHarvested),
            innaiHarvestedUSD: Number(innaiHarvestedUSD)
         }));
    }
};

const stakedValue = {
    properties: [
        'id',
        'liquidityTokenBalance',
        'pair { id, totalSupply, reserveETH, reserveUSD }'
    ],

    callback(results) {
        return ({
            id: results.id,
            liquidityTokenBalance: Number(results.liquidityTokenBalance),
            totalSupply: Number(results.pair.totalSupply),
            totalValueETH: Number(results.pair.reserveETH),
            totalValueUSD: Number(results.pair.reserveUSD)
        })
    }
};

const user = {
    properties: [
        'id',
        'address',
        'pool { id, pair, balance, accInnovaiPerShare, lastRewardBlock }',
        'amount',
        'rewardDebt',
        'entryUSD',
        'exitUSD',
        'innaiHarvested',
        'innaiHarvestedUSD',
    ],

    callback(results) {
        return results.map(entry => ({
            id: entry.id,
            address: entry.address,
            poolId: Number(entry.id.split("-")[0]),
            pool: entry.pool ? {
                id: entry.pool.id,
                pair: entry.pool.pair,
                balance: Number(entry.pool.balance),
                accInnovaiPerShare: BigInt(entry.pool.accInnovaiPerShare),
                lastRewardBlock: Number(entry.pool.lastRewardBlock)
            } : undefined,
            amount: Number(entry.amount),
            rewardDebt: BigInt(entry.rewardDebt),
            entryUSD: Number(entry.entryUSD),
            exitUSD: Number(entry.exitUSD),
            innaiHarvested: Number(entry.innaiHarvested),
            innaiHarvestedUSD: Number(entry.innaiHarvestedUSD),
        }));
    }
};

const apys = {
    callback24h(results, results24h) {
        return results.map(result => {
            const result24h = results24h.find(e => e.id === result.id) || result;

            return ({
                ...result,

                innlpBalanceChange: (result.innlpBalance / result24h.innlpBalance) * 100 - 100,
                innlpBalanceChangeCount: result.innlpBalance - result24h.innlpBalance,

                userCountChange: (result.userCount / result24h.userCount) * 100 - 100,
                userCountChangeCount: result.userCount - result24h.userCount,

                innaiHarvestedChange: (result.innaiHarvested / result24h.innaiHarvested) * 100 - 100,
                innaiHarvestedChangeCount: result.innaiHarvested - result24h.innaiHarvested,
            });
        });
    }
}
